==> testcafe-13.1.3.tgz: package/lib/server/recorder.js <==
var fs = require("fs"), path = require("path"), url = require("url"), util = require("util"), uuid = require("node-uuid"), cmd = require("../shared/service_msg_cmd"), ERR = require("./server_errs"), errMsgBuilder = require("./err_msg_builder"), FixtureCode = require("../fixture_code/"), injector = require("./injector"), urlUtil = require("../shared/url_util");

var RECORDER_PAGE_HTML_TEMPLATE = "<!DOCTYPE html>" + "<html>" + "<head>" + '<meta http-equiv="Content-Type" content="text/html; charset=utf-8">' + "<title></title>" + '<link href="%s" rel="stylesheet" type="text/css">' + '<script type="text/javascript">%s</script> ' + "</head>" + "<body></body>" + "</html>";

var Recorder = exports.Recorder = function(proxy, vfs, cookieShelf, assetsManager) {
    this.proxy = proxy;
    this.vfs = vfs;
    this.cookieShelf = cookieShelf;
    this.assetsManager = assetsManager;
    this.silentMode = false;
    this.disableTooltip = false;
    this.recordings = {};
    var proxyEvents = this.proxy.events.for(Recorder.JOB_OWNER_TOKEN);
    proxyEvents.listen("authCredentialsRequested", this._onAuthCredentialsRequested.bind(this));
    proxyEvents.listen("beforeProxyResponse", this._onBeforeProxyResponse.bind(this));
    proxyEvents.listen("error", this._onProxyError.bind(this));
    var serviceChannelEvents = this.proxy.serviceChannel.events.for(Recorder.JOB_OWNER_TOKEN);
    serviceChannelEvents.listen("serviceMsg", this._onServiceMsg.bind(this));
    serviceChannelEvents.listen("testCafeScriptRequested", this._onTestCafeScriptRequested.bind(this));
};

Recorder.JOB_OWNER_TOKEN = "r";

Recorder.prototype._onAuthCredentialsRequested = function(proxyCtx, callback) {
    var recording = this.recordings[proxyCtx.jobInfo.uid];
    callback(recording ? recording.authCredentials : null);
};

Recorder.prototype._onBeforeProxyResponse = function(ctx, callback) {
    var recording = this.recordings[ctx.jobInfo.uid];
    if (!recording) {
        ctx.res.statusCode = 500;
        ctx.res.end();
        return;
    }
    if (ctx.originRes.statusCode === 401 && ctx.originResContentInfo.isPage) {
        this._redirectToAuthenticationPage(ctx, recording);
        return;
    }
    if (ctx.originRes.statusCode >= 400 && !ctx.isXhr && !ctx.isIFrame) {
        this._redirectToServerErrorPage(ctx, recording.returnUrl);
        return;
    }
    if (ctx.originResContentInfo.isPage && ctx.originRes.statusCode === 200) {
        this._injectRecording(ctx, recording, callback);
    } else {
        callback();
    }
};

Recorder.prototype._onProxyError = function(proxyCtx, err) {
    var recording = this.recordings[proxyCtx.jobInfo.uid];
    if (recording && (proxyCtx.originResContentInfo && proxyCtx.originResContentInfo.isPage || proxyCtx.isPageReqCandidate)) {
        this._redirectToServerErrorPage(proxyCtx, recording.returnUrl);
    }
};

Recorder.prototype._onServiceMsg = function(msg, callback) {
    var recording = this.recordings[msg.jobUid], res = null;
    if (recording) {
        switch (msg.cmd) {
          case cmd.STEPS_INFO_GET:
            res = recording.stepsInfo;
            break;

          case cmd.STEPS_INFO_SET:
            recording.stepsInfo = msg.stepsInfo;
            break;

          case cmd.SILENT_MODE_SET:
            this.silentMode = msg.silentMode;
            break;

          case cmd.TOOLBAR_POSITION_SET:
            recording.toolbarPosition = msg.toolbarPosition;
            break;

          case cmd.DISABLE_TOOLTIP_SET:
            this.disableTooltip = msg.disableTooltip;
            break;

          case cmd.RESTART_RECORD:
            var recorder = this;
            recorder.getFixtureUid(recording.fixturePath, recording.fixtureFileName, function(uid) {
                recorder.startRecording(uid, recording.returnUrl, function(err, recordingUrl) {
                    callback({
                        err: err ? errMsgBuilder.build(err) : err,
                        recordingUrl: recordingUrl
                    });
                });
            });
            return;

          case cmd.ADD_TEST_CODE:
            this.completeRecording(msg.jobUid, msg.testName, function(res) {
                var err = res.err, fixtureFilename = res.fixtureFileName;
                callback({
                    err: err ? errMsgBuilder.build(err) : err,
                    fixtureFilename: fixtureFilename
                });
            });
            return;

          case cmd.AUTH_CREDENTIALS_SET:
            var vfs = this.vfs;
            vfs.getRealPath(recording.fixturePath, recording.fixtureFileName, function(fileName) {
                var username = msg.username, password = msg.password;
                FixtureCode.CodeGenerator.editDirectives(fileName, {
                    auth: username + ":" + password
                }, function(err) {
                    if (err) {
                        callback(errMsgBuilder.build(err));
                        return;
                    }
                    recording.authCredentials = {
                        username: username,
                        password: password
                    };
                    vfs.forceBuild(callback);
                });
            });
            return;

          default:
            break;
        }
    }
    callback(res);
};

Recorder.prototype._onTestCafeScriptRequested = function(refererInfo, callback) {
    var recorder = this, recording = this.recordings[refererInfo.jobInfo.uid];
    this.cookieShelf.getClientCookieString(refererInfo.jobInfo, refererInfo.originResourceInfo.url, function(cookie) {
        var recordingScript = recorder.assetsManager.compileTemplate(recorder.assetsManager.TEMPLATES.RECORDING, {
            returnUrl: recording.returnUrl,
            jobUid: recording.uid,
            jobOwnerToken: Recorder.JOB_OWNER_TOKEN,
            serviceMsgUrl: recorder.proxy.serviceChannel.serviceMsgUrl,
            cookie: cookie.replace(/'/g, "\\'"),
            originHost: refererInfo.originResourceInfo.host,
            originProtocol: refererInfo.originResourceInfo.protocol,
            originHostname: refererInfo.originResourceInfo.hostname,
            originPort: refererInfo.originResourceInfo.port || "",
            silentMode: recorder.silentMode,
            disableTooltip: recorder.disableTooltip,
            toolbarPosLeft: recording.toolbarPosition.left,
            toolbarPosTop: recording.toolbarPosition.top
        });
        callback(recordingScript);
    });
};

Recorder.prototype._redirectToServerErrorPage = function(proxyCtx, returnUrl) {
    var parsedProxyUrl = urlUtil.parseProxyUrl(proxyCtx.req.url), errorPageScript = this.assetsManager.compileTemplate(this.assetsManager.TEMPLATES.RECORDING_ERR_PAGE, {
        returnUrl: returnUrl,
        errorCode: proxyCtx.originRes ? proxyCtx.originRes.statusCode : 500,
        originUrl: parsedProxyUrl ? url.format(parsedProxyUrl.originResourceInfo) : ""
    });
    proxyCtx.res.statusCode = 500;
    proxyCtx.res.setHeader("content-type", "text/html");
    proxyCtx.res.end(util.format(RECORDER_PAGE_HTML_TEMPLATE, this.assetsManager.uiStyleUrl, errorPageScript));
};

Recorder.prototype._redirectToAuthenticationPage = function(proxyCtx, recording) {
    var parsedProxyUrl = urlUtil.parseProxyUrl(proxyCtx.req.url), authenticationPageScript = this.assetsManager.compileTemplate(this.assetsManager.TEMPLATES.RECORDING_AUTHENTICATION_PAGE, {
        jobUid: recording.uid,
        jobOwnerToken: Recorder.JOB_OWNER_TOKEN,
        serviceMsgUrl: this.proxy.serviceChannel.serviceMsgUrl,
        credentials: recording.authCredentials ? recording.authCredentials.username + ":" + recording.authCredentials.password : "",
        returnUrl: recording.returnUrl,
        targetUrl: proxyCtx.req.url,
        originUrl: parsedProxyUrl ? url.format(parsedProxyUrl.originResourceInfo) : ""
    });
    var html = util.format(RECORDER_PAGE_HTML_TEMPLATE, this.assetsManager.uiStyleUrl, authenticationPageScript);
    proxyCtx.res.statusCode = 401;
    proxyCtx.res.setHeader("content-type", "text/html");
    proxyCtx.res.end(html);
};

Recorder.prototype._injectRecording = function(proxyCtx, recording, callback) {
    var recorder = this, opt = {
        urlReplacer: this.proxy.getResourceUrlReplacer(proxyCtx),
        styleUrl: this.assetsManager.uiStyleUrl,
        scriptUrl: this.proxy.serviceChannel.testCafeScriptUrl
    };
    injector.injectInPage(proxyCtx.originResBodyChunks, proxyCtx.originResContentInfo.encoding, proxyCtx.originResContentInfo.charset, opt, function(err, html) {
        if (err) {
            proxyCtx.originRes.statusCode = 500;
            recorder._redirectToServerErrorPage(proxyCtx, recording.returnUrl);
        } else {
            proxyCtx.originResBodyChunks = [ html ];
            callback();
        }
    });
};

Recorder.prototype.startRecording = function(fixtureUid, returnUrl, callback) {
    var recorder = this;
    this.vfs.getSuite(function(suite) {
        var fixture = suite.fixtures[fixtureUid];
        if (!fixture) {
            callback({
                code: ERR.RECORDER_TARGET_FIXTURE_DOESNT_EXISTS
            });
            return;
        }
        var recordingUid = uuid.v4();
        recorder.recordings[recordingUid] = {
            authCredentials: fixture.authCredentials,
            fixtureFileName: fixture.fileName,
            fixturePath: fixture.path,
            returnUrl: returnUrl,
            stepsInfo: [],
            uid: recordingUid,
            toolbarPosition: {
                left: null,
                top: null
            }
        };
        try {
            var recordingUrl = recorder.proxy.getProxyUrl(fixture.page, recordingUid, Recorder.JOB_OWNER_TOKEN);
        } catch (err) {
            callback(err);
            return;
        }
        callback(null, recordingUrl);
    });
};

Recorder.prototype.getFixtureUid = function(fixturePath, filename, callback) {
    this.vfs.getSuite(function(suite) {
        Object.keys(suite.fixtures).forEach(function(uid) {
            if (suite.fixtures[uid].fileName === filename && suite.fixtures[uid].path.join("/") === fixturePath.join("/")) {
                callback(uid);
            }
        });
        callback(null);
    });
};

Recorder.prototype.completeRecording = function(recordingUid, testName, callback) {
    var recording = this.recordings[recordingUid], vfs = this.vfs;
    if (!recording) {
        callback({
            err: {
                code: ERR.RECORDER_RECORDING_DOESNT_EXISTS
            }
        });
        return;
    }
    var recorder = this;
    this.cookieShelf.removeCookies({
        ownerToken: Recorder.JOB_OWNER_TOKEN,
        uid: recordingUid
    });
    vfs.getRealPath(recording.fixturePath, recording.fixtureFileName, function(fileName) {
        if (!recording.stepsInfo.length) {
            callback({
                err: {
                    code: ERR.RECORDER_RECORDING_IS_EMPTY
                }
            });
            return;
        }
        if (!/[\S]+/.test(testName)) {
            callback({
                err: {
                    code: ERR.RECORDER_TEST_NAME_IS_EMPTY
                }
            });
            return;
        }
        vfs.removeFSWatchers();
        FixtureCode.CodeGenerator.addTest(fileName, testName, recording.stepsInfo, function(err) {
            if (!err && recorder.recordings[recordingUid]) {
                delete recorder.recordings[recordingUid];
            }
            vfs.forceBuild(function() {
                callback({
                    err: err,
                    fixtureFileName: recording.fixtureFileName
                });
            });
        });
    });
};

==> testcafe-14.1.3.tgz: package/lib/server/recorder.js <==
var util = require("util"), uuid = require("node-uuid"), cmd = require("../shared/service_msg_cmd"), ERR = require("./server_errs"), errMsgBuilder = require("./err_msg_builder"), FixtureCode = require("../fixture_code/"), AssetsManager = require("./assets_manager");

var RECORDER_PAGE_HTML_TEMPLATE = "<!DOCTYPE html>" + "<html>" + "<head>" + '<meta http-equiv="Content-Type" content="text/html; charset=utf-8">' + "<title></title>" + '<link href="%s" rel="stylesheet" type="text/css">' + '<script type="text/javascript" src="%s"></script> ' + '<script type="text/javascript" src="%s"></script> ' + '<script type="text/javascript">%s</script> ' + '<script type="text/javascript">%s</script> ' + "</head>" + "<body></body>" + "</html>";

var Recorder = exports.Recorder = function(hammerhead, vfs) {
    this.hammerhead = hammerhead;
    this.vfs = vfs;
    this.silentMode = false;
    this.disableTooltip = false;
    this.recordings = {};
    var proxyEvents = hammerhead.proxyEvents.for(Recorder.JOB_OWNER_TOKEN), serviceEvents = hammerhead.serviceEvents.for(Recorder.JOB_OWNER_TOKEN);
    proxyEvents.listen("authCredentialsRequested", this._onAuthCredentialsRequested.bind(this));
    proxyEvents.listen("originResponse", this._onOriginResponse.bind(this));
    proxyEvents.listen("injectionError", this._onInjectionError.bind(this));
    proxyEvents.listen("error", this._onProxyError.bind(this));
    serviceEvents.listen("serviceMsg", this._onServiceMsg.bind(this));
    serviceEvents.listen("taskScriptRequested", this._onTaskScriptRequest.bind(this));
};

Recorder.JOB_OWNER_TOKEN = "r";

Recorder.prototype._onAuthCredentialsRequested = function(proxyCtx, callback) {
    var recording = this.recordings[proxyCtx.jobInfo.uid];
    callback(recording ? recording.authCredentials : null);
};

Recorder.prototype._onOriginResponse = function(ctx, callback) {
    var recording = this.recordings[ctx.jobInfo.uid];
    if (!recording) {
        ctx.res.statusCode = 500;
        ctx.res.end();
        return;
    }
    if (ctx.originRes.statusCode === 401 && ctx.originResContentInfo.isPage) {
        this._redirectToAuthenticationPage(ctx, recording);
        return;
    }
    if (ctx.originRes.statusCode >= 400 && !ctx.isXhr && !ctx.isIFrame) {
        this._redirectToServerErrorPage(ctx, recording.returnUrl);
        return;
    }
    callback();
};

Recorder.prototype._onInjectionError = function(ctx) {
    var recording = this.recordings[ctx.jobInfo.uid];
    ctx.originRes.statusCode = 500;
    this._redirectToServerErrorPage(ctx, recording.returnUrl);
};

Recorder.prototype._onProxyError = function(proxyCtx) {
    var recording = this.recordings[proxyCtx.jobInfo.uid];
    if (recording && (proxyCtx.originResContentInfo && proxyCtx.originResContentInfo.isPage || proxyCtx.isPageReqCandidate)) {
        this._redirectToServerErrorPage(proxyCtx, recording.returnUrl);
    }
};

Recorder.prototype._onServiceMsg = function(msg, callback) {
    var recording = this.recordings[msg.jobUid], res = null;
    if (recording) {
        switch (msg.cmd) {
          case cmd.STEPS_INFO_GET:
            res = recording.stepsInfo;
            break;

          case cmd.STEPS_INFO_SET:
            recording.stepsInfo = msg.stepsInfo;
            break;

          case cmd.SILENT_MODE_SET:
            this.silentMode = msg.silentMode;
            break;

          case cmd.TOOLBAR_POSITION_SET:
            recording.toolbarPosition = msg.toolbarPosition;
            break;

          case cmd.DISABLE_TOOLTIP_SET:
            this.disableTooltip = msg.disableTooltip;
            break;

          case cmd.RESTART_RECORD:
            var recorder = this;
            recorder.getFixtureUid(recording.fixturePath, recording.fixtureFileName, function(uid) {
                recorder.startRecording(uid, recording.returnUrl, function(err, recordingUrl) {
                    callback({
                        err: err ? errMsgBuilder.build(err) : err,
                        recordingUrl: recordingUrl
                    });
                });
            });
            return;

          case cmd.ADD_TEST_CODE:
            this.completeRecording(msg.jobUid, msg.testName, function(res) {
                var err = res.err, fixtureFilename = res.fixtureFileName;
                callback({
                    err: err ? errMsgBuilder.build(err) : err,
                    fixtureFilename: fixtureFilename
                });
            });
            return;

          case cmd.AUTH_CREDENTIALS_SET:
            var vfs = this.vfs;
            vfs.getRealPath(recording.fixturePath, recording.fixtureFileName, function(fileName) {
                var username = msg.username, password = msg.password;
                FixtureCode.CodeGenerator.editDirectives(fileName, {
                    auth: username + ":" + password
                }, function(err) {
                    if (err) {
                        callback(errMsgBuilder.build(err));
                        return;
                    }
                    recording.authCredentials = {
                        username: username,
                        password: password
                    };
                    vfs.forceBuild(callback);
                });
            });
            return;

          default:
            break;
        }
    }
    callback(res);
};

Recorder.prototype._onTaskScriptRequest = function(refererInfo, cookie, callback) {
    var recorder = this, recording = this.recordings[refererInfo.jobInfo.uid];
    var recordingScript = AssetsManager.compileTemplate(AssetsManager.TEMPLATES.RECORDING, {
        returnUrl: recording.returnUrl,
        jobUid: recording.uid,
        jobOwnerToken: Recorder.JOB_OWNER_TOKEN,
        cookie: cookie.replace(/'/g, "\\'"),
        originHost: refererInfo.originResourceInfo.host,
        originProtocol: refererInfo.originResourceInfo.protocol,
        originHostname: refererInfo.originResourceInfo.hostname,
        originPort: refererInfo.originResourceInfo.port || "",
        silentMode: recorder.silentMode,
        disableTooltip: recorder.disableTooltip,
        toolbarPosLeft: recording.toolbarPosition.left,
        toolbarPosTop: recording.toolbarPosition.top
    });
    callback(recordingScript);
};

Recorder.prototype._redirectToServerErrorPage = function(proxyCtx, returnUrl) {
    var errorPageScript = AssetsManager.compileTemplate(AssetsManager.TEMPLATES.RECORDING_ERR_PAGE, {
        returnUrl: returnUrl,
        errorCode: proxyCtx.originRes ? proxyCtx.originRes.statusCode : 500,
        originUrl: proxyCtx.originResourceInfo.url
    });
    var html = util.format(RECORDER_PAGE_HTML_TEMPLATE, this.hammerhead.staticCssUrl, this.hammerhead.hammerheadScriptUrl, this.hammerhead.staticScriptUrl, this.hammerhead.getSettingsScript(proxyCtx), errorPageScript);
    proxyCtx.res.statusCode = 500;
    proxyCtx.res.setHeader("content-type", "text/html");
    proxyCtx.res.end(html);
};

Recorder.prototype._redirectToAuthenticationPage = function(proxyCtx, recording) {
    var authenticationPageScript = AssetsManager.compileTemplate(AssetsManager.TEMPLATES.RECORDING_AUTHENTICATION_PAGE, {
        jobUid: recording.uid,
        jobOwnerToken: Recorder.JOB_OWNER_TOKEN,
        credentials: recording.authCredentials ? recording.authCredentials.username + ":" + recording.authCredentials.password : "",
        returnUrl: recording.returnUrl,
        targetUrl: proxyCtx.req.url,
        originUrl: proxyCtx.originResourceInfo.url
    });
    var html = util.format(RECORDER_PAGE_HTML_TEMPLATE, this.hammerhead.staticCssUrl, this.hammerhead.hammerheadScriptUrl, this.hammerhead.staticScriptUrl, this.hammerhead.getSettingsScript(proxyCtx), authenticationPageScript);
    proxyCtx.res.statusCode = 401;
    proxyCtx.res.setHeader("content-type", "text/html");
    proxyCtx.res.end(html);
};

Recorder.prototype.startRecording = function(fixtureUid, returnUrl, callback) {
    var recorder = this;
    this.vfs.getSuite(function(suite) {
        var fixture = suite.fixtures[fixtureUid];
        if (!fixture) {
            callback({
                code: ERR.RECORDER_TARGET_FIXTURE_DOESNT_EXISTS
            });
            return;
        }
        var recordingUid = uuid.v4();
        recorder.recordings[recordingUid] = {
            authCredentials: fixture.authCredentials,
            fixtureFileName: fixture.fileName,
            fixturePath: fixture.path,
            returnUrl: returnUrl,
            stepsInfo: [],
            uid: recordingUid,
            toolbarPosition: {
                left: null,
                top: null
            }
        };
        var recordingUrl = null;
        try {
            recordingUrl = recorder.hammerhead.getProxyUrl(fixture.page, recordingUid, Recorder.JOB_OWNER_TOKEN);
        } catch (err) {
            callback(err);
            return;
        }
        callback(null, recordingUrl);
    });
};

Recorder.prototype.getFixtureUid = function(fixturePath, filename, callback) {
    this.vfs.getSuite(function(suite) {
        Object.keys(suite.fixtures).forEach(function(uid) {
            if (suite.fixtures[uid].fileName === filename && suite.fixtures[uid].path.join("/") === fixturePath.join("/")) {
                callback(uid);
            }
        });
        callback(null);
    });
};

Recorder.prototype.completeRecording = function(recordingUid, testName, callback) {
    var recording = this.recordings[recordingUid], vfs = this.vfs;
    if (!recording) {
        callback({
            err: {
                code: ERR.RECORDER_RECORDING_DOESNT_EXISTS
            }
        });
        return;
    }
    var recorder = this;
    this.hammerhead.removeCookies({
        ownerToken: Recorder.JOB_OWNER_TOKEN,
        uid: recordingUid
    });
    vfs.getRealPath(recording.fixturePath, recording.fixtureFileName, function(fileName) {
        if (!recording.stepsInfo.length) {
            callback({
                err: {
                    code: ERR.RECORDER_RECORDING_IS_EMPTY
                }
            });
            return;
        }
        if (!/[\S]+/.test(testName)) {
            callback({
                err: {
                    code: ERR.RECORDER_TEST_NAME_IS_EMPTY
                }
            });
            return;
        }
        vfs.removeFSWatchers();
        FixtureCode.CodeGenerator.addTest(fileName, testName, recording.stepsInfo, function(err) {
            if (!err && recorder.recordings[recordingUid]) {
                delete recorder.recordings[recordingUid];
            }
            vfs.forceBuild(function() {
                callback({
                    err: err,
                    fixtureFileName: recording.fixtureFileName
                });
            });
        });
    });
};
